/*  actest.c - a Geany plugin
 *
 *  Copyright 2009 /me
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 */


#include "geany.h"
#include "support.h"

#include <ctype.h>
#include <string.h>
#ifdef HAVE_LOCALE_H
# include <locale.h>
#endif

#include "utils.h"
#include "sciwrappers.h"
#include "editor.h"
#include "document.h"
#include "keybindings.h"

#include "plugindata.h"
#include "geanyfunctions.h"

#include <gdk/gdkkeysyms.h>

GeanyPlugin     *geany_plugin;
GeanyData       *geany_data;
GeanyFunctions  *geany_functions;

PLUGIN_VERSION_CHECK(140);
PLUGIN_SET_INFO(_("AutoComplete Test"), _("blub"), "0.1", "/me");


/* Keybinding(s) */
enum
{
	KB_AC_TEST,
	KB_COUNT
};
PLUGIN_KEY_GROUP(actest, KB_COUNT)


static gboolean editor_notify_cb(GObject *object, GeanyEditor *editor,
								 SCNotification *nt, gpointer data)
{
	switch (nt->nmhdr.code)
	{
		case SCN_USERLISTSELECTION:
		{
			if (nt->listType == 3)
			{
				gint start = nt->lParam - 1;

				while (start > 0 &&
					strchr(GEANY_WORDCHARS, sci_get_char_at(editor->sci, start)) != NULL)
				{
					start--;
				}
				start++;

				sci_set_selection_start(editor->sci, start);
				sci_set_selection_end(editor->sci, nt->lParam);
				sci_replace_sel(editor->sci, "");	/* clear root of word */

				sci_insert_text(editor->sci, start, nt->text);
				sci_set_current_position(editor->sci, start + strlen(nt->text), TRUE);

				return TRUE;
			}
			break;
		}
	}

	return FALSE;
}


/* Algorithm based on based on Scite's StartAutoCompleteWord() */
static void ac_test_complete_word(void)
{
	GeanyDocument *doc = document_get_current();
	ScintillaObject *sci;
	gchar *linebuf, *root, *word;
	gint pos, line, len, startword, current, word_end;
	gint pos_find, flags;
	guint word_length, rootlen;
	GString *words;
	struct TextToFind ttf;

	if (doc == NULL)
		return;

	sci = doc->editor->sci;
	pos = sci_get_current_position(sci);
	line = sci_get_line_from_position(sci, pos);
	current = pos - sci_get_position_from_line(sci, line);
	startword = current;
	linebuf = sci_get_line(sci, line);
	len = sci_get_length(sci);

	/* find the start of the current word */
	while ((startword > 0) && (strchr(GEANY_WORDCHARS, linebuf[startword - 1])))
		startword--;
	linebuf[current] = '\0';
	root = linebuf + startword;
	rootlen = current - startword;
	current = pos - rootlen;

	if (rootlen == 0)
	{
		g_free(linebuf);
		return;
	}

	ttf.lpstrText = root;
	ttf.chrg.cpMin = 0;
	ttf.chrg.cpMax = len;
	ttf.chrgText.cpMin = 0;
	ttf.chrgText.cpMax = 0;
	flags = SCFIND_WORDSTART | SCFIND_MATCHCASE;

	words = g_string_sized_new(256);
	g_string_append_c(words, ' ');

	/* search the whole document for the word root and collect results */
	pos_find = scintilla_send_message(sci, SCI_FINDTEXT, flags, (uptr_t) &ttf);
	while (pos_find >= 0 && pos_find < len)
	{
		word_end = pos_find + rootlen;
		if (pos_find != current)
		{
			while (word_end < len && strchr(GEANY_WORDCHARS, sci_get_char_at(sci, word_end)) != NULL)
				word_end++;

			word_length = word_end - pos_find;
			if (word_length > rootlen)
			{
				word = g_malloc0(word_length + 3);
				sci_get_text_range(sci, pos_find, word_end, word + 1);
				word[0] = ' ';
				word[word_length + 1] = ' ';
				/* search the words string whether we already have the word in, otherwise add it */
				if (strstr(words->str, word) == NULL)
					g_string_append(words, word + 1);
				g_free(word);
			}
		}
		ttf.chrg.cpMin = word_end;
		pos_find = scintilla_send_message(sci, SCI_FINDTEXT, flags, (uptr_t) &ttf);
	}

	if (words->len > 1)
	{
		g_strdelimit(words->str, " ", '\n');
		words->str[words->len - 1] = '\0'; /* remove the trailing '\n' */
		/* FIXME Geany uses list type 0 (auto completion), 1 and 2(maybe?), so we use 3 */
		scintilla_send_message(sci, SCI_USERLISTSHOW, 3, (sptr_t) words->str + 1);
	}
	else
		scintilla_send_message(sci, SCI_AUTOCCANCEL, 0, 0);

	g_free(linebuf);
	g_string_free(words, TRUE);
}


static void kb_ac_test(G_GNUC_UNUSED guint key_id)
{
	ac_test_complete_word();
}


void plugin_init(G_GNUC_UNUSED GeanyData *data)
{
	/* init gettext and friends */
	/*main_locale_init(LOCALEDIR, GETTEXT_PACKAGE);*/

	keybindings_set_item(plugin_key_group, KB_AC_TEST, kb_ac_test,
		GDK_less, GDK_CONTROL_MASK, "ac_test", _("AC Test"), NULL);
}

void plugin_cleanup(void)
{
}


PluginCallback plugin_callbacks[] =
{
	{ "editor-notify", (GCallback) &editor_notify_cb, TRUE, NULL },
	{ NULL, NULL, FALSE, NULL }
};

